# -*- coding: utf-8 -*-
# Copyright 2021 Cardiff University
# Distributed under the terms of the BSD-3-Clause license

"""Tests for :mod:`igwn_auth_utils`.
"""

__author__ = "Duncan Macleod <duncan.macleod@ligo.org>"
